/*
  date_time_repeater.hpp

  Copyright (C) 2011 Taiji Yamada <taiji@aihara.co.jp>
 */
#ifndef _DATE_TIME_REPEATER_HPP_
#define _DATE_TIME_REPEATER_HPP_
#include <boost/date_time.hpp>
#include "repeater.hpp"

inline
std::ostream &operator<<(std::ostream &os, const boost::gregorian::years &a)
{
  return os << a.number_of_years().as_number();
}

inline
std::istream &operator>>(std::istream &is, boost::gregorian::years &a)
{
  int y;
  is >> y;
  a = boost::gregorian::years(y);
  return is;
}

#define T boost::gregorian::date
#define D boost::gregorian::years
template<>
struct repeater<T, D> {
  T front, back;
  D diff;
  repeater() : front(), back(), diff(0) {}
  repeater(T s) : front(s), back(), diff(0) {}
  repeater(T s, T e) : front(s), back(e), diff(1) {}
  repeater(T s, D d, T e) : front(s), back(e), diff(d) {}
  const bool is_repeat() const { return (diff != D(0)); }
  const bool is_unlimited() const { return (is_repeat() && (front == back)); }
  const long count() const { return long((!is_repeat()) ? 1 : (!is_unlimited()) ? ((back - front)/boost::gregorian::date_duration(diff.number_of_years().as_number()*365.25) + 1) : 0); }
  const T last() const { return (!is_repeat()) ? front : (!is_unlimited()) ? front + diff*(count() - 1) : T(); }
  const boost::optional<T> before(const T &a) const
  {
    return (!(a < front)) ? boost::optional<T>((!is_repeat()) ? front : (last() < a) ? last() :
					       front + diff*((a - front)/boost::gregorian::date_duration(diff.number_of_years().as_number()*365.25))) : boost::optional<T>();
  }
  const boost::optional<T> after(const T &a) const
  {
    return (!(last() < a)) ? boost::optional<T>((!is_repeat()) ? front : (a < front) ? front :
						(is_unlimited()) ? front + diff*((a - front + boost::gregorian::date_duration(diff.number_of_years().as_number()*365.25 - 1))/boost::gregorian::date_duration(diff.number_of_years().as_number()*365.25)) :
						last() - diff*((last() - a)/boost::gregorian::date_duration(diff.number_of_years().as_number()*365.25))) : boost::optional<T>();
  }
};

/*
  0,0,*:	0		is_repeat() == false
  0,1,0:	0~		is_repeat() && is_unlimited() == true
  0,1,!0:	0~!0		is_repeat() == true
  0,2,0:	0~2~		is_repeat() && is_unlimited() == true
  0,2,!0:	0~2~!0		is_repeat() == true
 */
inline
std::ostream &operator<<(std::ostream &os, const repeater<T, D> &a)
{
  const char separator = '~';
  os << a.front;
  if (a.diff != D(0))
    if (a.diff == D(1)) {
      os << separator;
      if (a.front != a.back)
	os << a.back;
    }
    else {
      os << separator << a.diff << separator;
      if (a.front != a.back)
	os << a.back;
    }
  return os;
}

inline
std::istream &operator>>(std::istream &is, repeater<T, D> &a)
{
  const std::string separators(".~"), spaces("\t\r\n ");
  int i = 0, j = 0;
  std::stringstream ss[3];
  char c;
  a.front = a.back = T();
  a.diff = D(0);
  do {
    is >> c;
    if (!is.good()) { is.clear(); break; }
    if (separators.find(c) != std::string::npos) ++i;
    else if (spaces.find(c) != std::string::npos) ;
    else ss[j=i] << c;
  } while (j < 3);
  ++j;
  ss[0] >> a.front;
  if (i != j) {
    if (j == 2) {
      ss[1] >> a.back;
      a.diff = D(1);
    }
    else {
      ss[1] >> a.diff;
      ss[2] >> a.back;
    }
  }
  else {
    if (j == 1)
      a.diff = D(1);
    else
      ss[1] >> a.diff;
    a.back = a.front;
  }
  return is;
}
#undef T
#undef D

#endif
